import warnings
from typing import Callable, Protocol, runtime_checkable

import numpy as np
import pandas as pd

from daindex.core import deterioration_index


@runtime_checkable
class ProbabilisticModel(Protocol):
    def predict_proba(self, X: np.ndarray) -> np.ndarray: ...


def obtain_da_index(
    df,
    cohort_name,
    scores,
    score_bounds: tuple[float, float],
    det_feature: str,
    det_threshold: float,
    det_label: str,
    det_feature_func: Callable,
    det_list_lengths: list[int],
    min_det_v: float,
    max_det_v: float,
    is_discrete: bool,
    reverse: bool,
    optimise_bandwidth: bool,
) -> tuple[int, float]:
    """
    Calculates the Deterioration Allocation Index (DAI) for a given cohort.

    Args:
        df: The DataFrame containing the data.
        cohort_name: The name of the cohort.
        scores: The list of scores corresponding to the DataFrame rows.
        score_bounds: The lower and upper bounds for the scores to be considered.
        det_feature: The feature to be used for DAI calculation.
        det_threshold: The threshold value for the deterioration index.
        det_label: The label for the deterioration index.
        det_feature_func: A function to apply to each row to extract the feature value.
        det_list_lengths: A list of acceptable lengths for the det_list, in descending order.
        min_det_v: The minimum value for the deterioration index.
        max_det_v: The maximum value for the deterioration index.
        is_discrete: Whether the feature is discrete.
        reverse: Whether to reverse the order of the feature values.
        optimise_bandwidth: Whether to search for the optimal bandwidth.

    Returns:
        A tuple containing:
            - int: The length of the det_list.
            - float: The k-step value from the deterioration index calculation.

    Raises:
        UserWarning: If the number of samples is sub-optimal or insufficient for DAI calculation.

    """
    # sort det_list_lengths into descending order
    det_list_lengths.sort(reverse=True)
    lb, ub = score_bounds

    det_list = []
    i = 0
    for idx, r in df.iterrows():
        p = scores[i]
        if lb <= p <= ub:
            if det_feature_func is not None:
                det_list.append(det_feature_func(r))
            else:
                det_list.append(r[det_feature])
        i += 1
    for det_list_length in det_list_lengths:
        if len(det_list) >= det_list_length:
            if det_list_length != det_list_lengths[0]:
                warnings.warn(
                    f"Sub-optimal number of samples for DAI calculation, {len(det_list)} is acceptable but {det_list_lengths[0]} is preferred."
                )
            break
    else:
        warnings.warn(f"Insufficient number of samples for DAI calculation, {len(det_list)} < {det_list_lengths[-1]}.")
        return len(det_list), 0

    X = np.array(det_list)
    di_ret = deterioration_index(
        X[~np.isnan(X)].reshape(-1, 1),
        min_det_v,
        max_det_v,
        threshold=det_threshold,
        plot_title=f"{cohort_name} | {det_label}",
        is_discrete=is_discrete,
        reverse=reverse,
        optimise_bandwidth=optimise_bandwidth,
        do_plot=False,
    )
    return len(det_list), di_ret["k-step"]


def get_scores(models: list[ProbabilisticModel], df: pd.DataFrame, feature_list: list[str]) -> np.ndarray:
    """
    Computes the mean predicted probabilities for a list of models.

    Args:
        models: A list of trained model objects that have a `predict_proba` method.
        df: A DataFrame containing the data to be used for predictions.
        feature_list: A list of column names in `df` to be used as features for prediction.

    Returns:
        An array of mean predicted probabilities for the positive class.
    """
    predicted_probs = np.array([m.predict_proba(df[feature_list].to_numpy()) for m in models])
    return predicted_probs[:, :, 1].mean(axis=0)


def get_da_values_on_models(
    df: pd.DataFrame,
    cohort_name: str,
    model: list[ProbabilisticModel] | ProbabilisticModel,
    feature_list: list[str],
    det_feature: str,
    det_threshold: float,
    det_label: str = None,
    det_feature_func: Callable = None,
    det_list_lengths: list[int] = [20, 10, 5],
    steps: int = 50,
    is_discrete: bool = False,
    reverse: bool = False,
    optimise_bandwidth: bool = True,
    score_margin_multiplier: float = 5.0,
) -> list[tuple[int, float]]:
    """
    Compute the deterioration index on the predictions of the model.

    Args:
        df: The DataFrame containing the data.
        cohort_name: The name of the cohort being analyzed.
        model: The predictive model or list of models to be used.
        feature_list: List of feature names to be used for prediction.
        det_feature: The feature name representing the deterioration metric.
        det_threshold: The threshold value for determining deterioration.
        det_label: The label for the deterioration index. Defaults to None to be replaced by `det_feature`.
        det_feature_func: The function to be used to extract the deterioration feature. Defaults to None as we assume it is in the `df`.
        det_list_lengths: The list of acceptable lengths to be used for the deterioration index calculation. The first element is the preferred minimum length. Defaults to [20, 10, 5].
        steps: The number of steps for the deterioration index calculation. Defaults to 50.
        is_discrete: Flag indicating if the deterioration feature is discrete. Defaults to False.
        reverse: Flag indicating if the deterioration index should be reversed. Defaults to False.
        optimise_bandwidth: Flag indicating if the bandwidth of the kde should be searched. Defaults to True.
        score_margin_multiplier: The multiplier to be used for the score margin, 1.0 results in no overlap, some overlap works well, essentially this smooths. Defaults to 5.0.

    Returns:
        A list of deterioration index values computed at each step,
        the length of the list of scores within the score range (there is a small upper and lower bound applied)
        and finally the da index value itself.
    """

    det_label = det_label or det_feature
    models = model if isinstance(model, list) else [model]
    ret = []

    min_det_v = np.min(df[det_feature])
    max_det_v = np.max(df[det_feature])

    scores = get_scores(models, df, feature_list)
    step_scores = np.linspace(0, 1, steps + 1)[1:] - 1 / (2 * steps)

    for s in step_scores:
        step_score_bounds = (
            max(0.0, s - (1 / (2 * steps)) * score_margin_multiplier),
            min(1.0, s + (1 / (2 * steps)) * score_margin_multiplier),
        )
        ret.append(
            (
                s,
                *obtain_da_index(
                    df,
                    cohort_name,
                    scores,
                    step_score_bounds,
                    det_feature,
                    det_threshold,
                    det_label,
                    det_feature_func,
                    det_list_lengths,
                    min_det_v,
                    max_det_v,
                    is_discrete,
                    reverse,
                    optimise_bandwidth,
                ),
            )
        )

    return ret


def get_da_values_on_predictions(
    df: pd.DataFrame,
    cohort_name: str,
    preds_col: str,
    det_feature: str,
    det_threshold: float,
    det_label: str = None,
    det_feature_func: Callable = None,
    det_list_lengths: list[int] = [20, 10, 5],
    steps: int = 50,
    is_discrete: bool = False,
    reverse: bool = False,
    optimise_bandwidth: bool = True,
    score_margin_multiplier: float = 5.0,
) -> list[tuple[int, float]]:
    """
    Compute the deterioration index on the predictions of the model.

    Args:
        df: The DataFrame containing the data.
        cohort_name: The name of the cohort being analyzed.
        preds_col: The column name containing the predictions.
        det_feature: The feature name representing the deterioration metric.
        det_threshold: The threshold value for determining deterioration.
        det_label: The label for the deterioration index. Defaults to None to be replaced by `det_feature`.
        det_feature_func: The function to be used to extract the deterioration feature. Defaults to None as we assume it is in the `df`.
        det_list_lengths: The list of acceptable lengths to be used for the deterioration index calculation. The first element is the preferred minimum length. Defaults to [20, 10, 5].
        steps: The number of steps for the deterioration index calculation. Defaults to 50.
        is_discrete: Flag indicating if the deterioration feature is discrete. Defaults to False.
        reverse: Flag indicating if the deterioration index should be reversed. Defaults to False.
        optimise_bandwidth: Flag indicating if the bandwidth of the kde should be searched. Defaults to True.
        score_margin_multiplier: The multiplier to be used for the score margin, 1.0 results in no overlap, some overlap works well, essentially this smooths. Defaults to 1.5.

    Returns:
        A list of deterioration index values computed at each step,
        the length of the list of scores within the score range (there is a small upper and lower bound applied)
        and finally the da index value itself.
    """

    det_label = det_label or det_feature
    ret = []

    min_det_v = np.min(df[det_feature])
    max_det_v = np.max(df[det_feature])

    scores = df[preds_col].values
    step_scores = np.linspace(0, 1, steps + 1)[1:] - 1 / (2 * steps)

    for s in step_scores:
        step_score_bounds = (
            max(0.0, s - (1 / (2 * steps)) * score_margin_multiplier),
            min(1.0, s + (1 / (2 * steps)) * score_margin_multiplier),
        )
        ret.append(
            (
                s,
                *obtain_da_index(
                    df,
                    cohort_name,
                    scores,
                    step_score_bounds,
                    det_feature,
                    det_threshold,
                    det_label,
                    det_feature_func,
                    det_list_lengths,
                    min_det_v,
                    max_det_v,
                    is_discrete,
                    reverse,
                    optimise_bandwidth,
                ),
            )
        )

    return ret



import logging
from typing import Any, Literal, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from sklearn.model_selection import GridSearchCV, KFold
from sklearn.neighbors import KernelDensity


def vis_plot(
    X: np.ndarray,
    x_d: np.ndarray,
    logprobs: np.ndarray,
    ylabel: str = "Percentage/PDF",
    title: str = "",
    vlines: list[float] = [],
    vline_colors: list[str] = [],
    vline_labels: list[str] = [],
    hist: bool = True,
    reverse: bool = False,
) -> None:
    """
    plot probability density function (PDF) and histogram for comparison
    """
    if hist:
        pd.DataFrame(X).rename(columns={0: "Histogram"}).plot.hist(
            bins=20, alpha=0.5, color="goldenrod", **{"density": True}, figsize=(18, 16)
        )
    else:
        plt.figure(figsize=(8, 2), dpi=100)
    plt.fill_between(x_d, np.exp(logprobs), alpha=0.5, label="Probability Density Function")
    # plt.plot(x, np.full_like(x, -0.005), '|k', markeredgewidth=1)
    for vl, c, label in zip(vlines, vline_colors, vline_labels):
        plt.axvline(x=vl, color=c, linestyle="--", label=f"{label}:{vl}")

    if len(vlines) > 0:
        if not reverse:
            plt.axvspan(vlines[-1], x_d[-1], facecolor="b", alpha=0.05)
        else:
            plt.axvspan(x_d[0], vlines[-1], facecolor="b", alpha=0.05)

    plt.legend()
    plt.ylabel(ylabel)
    plt.title(title)
    plt.yticks([])
    # plt.ylim((0, .03))
    # plt.xlim((10, 40))


def gridsearch_bandwidth(X: np.ndarray) -> float:
    """
    Search for the best bandwith for the KDE
    """
    bandwidths = np.linspace(0, 1, 20)
    grid = GridSearchCV(KernelDensity(kernel="gaussian"), {"bandwidth": bandwidths}, cv=KFold(5))
    grid.fit(X)
    return grid.best_params_["bandwidth"]


def get_threshold_index(
    threshold: float, low_bound: float, is_discrete: bool, prev_val_offset: float, step: float, boundary_offset: float
) -> int:
    """
    Calculate threshold value index
    """
    threshold_index = int((threshold - low_bound) / step)
    if threshold == low_bound + boundary_offset:
        logging.info(f"adjust boundary {threshold}")
        threshold_index = int((threshold - low_bound - boundary_offset) / step)
    elif is_discrete:
        # discrete values will lead to PDF shape like pulses,
        # it's important to start with the valley between the pulse you want to include and the one before that
        # the following is to find the valley index
        threshold_index_prev = int((threshold - prev_val_offset - low_bound) / step)
        threshold_index = int(threshold_index_prev + (threshold_index - threshold_index_prev) / 2)
    return threshold_index


def search_for_zero_mass_index(kde: KernelDensity, min_v: float, n_samples: int = 100) -> tuple[int, float]:
    """
    Search near zero probability mass for boundary adjustment
    """
    bins = np.linspace(min_v - 10, min_v, n_samples)
    pd_vals = np.exp(kde.score_samples(bins.reshape(-1, 1)))
    first_zero_idx = np.max(np.where(pd_vals < 1e-10))
    return first_zero_idx, bins[first_zero_idx]


def kde_estimate(
    X: np.ndarray,
    bandwidth: float | Literal["scott", "silverman"] = 1.0,
    kernel: Literal["gaussian", "tophat", "epanechnikov", "exponential", "linear", "cosine"] = "gaussian",
    optimise_bandwidth: bool = True,
) -> tuple[KernelDensity, float | Literal["scott", "silverman"]]:
    """
    Kernel density estimation to get probability
    """
    if optimise_bandwidth:
        bandwidth = gridsearch_bandwidth(X)
        logging.info(f"learned best bandwidth {bandwidth}")

    kde = KernelDensity(bandwidth=bandwidth, kernel=kernel)
    kde.fit(X)
    return kde, bandwidth


def deterioration_index(
    X: np.ndarray,
    low_bound: float,
    up_bound: float,
    threshold: float,
    n_samples: int = 10000,
    plot_title: str = "",
    is_discrete: bool = False,
    prev_discrete_value_offset=1,
    weight_sum_steps=10,
    reverse: bool = False,
    bandwidth: float | Literal["scott", "silverman"] = 1.0,
    kernel: Literal["gaussian", "tophat", "epanechnikov", "exponential", "linear", "cosine"] = "gaussian",
    optimise_bandwidth: bool = True,
    do_plot: bool = True,
) -> dict[str, int]:
    """
    Obtain deterioration index
    X - the random sample of measurements
    low_bound/up_nbound - the boundary values of the measurement
    n_samples - number of bins to use for probability calculation. default is 2000.
    plot_title - the title of the plot, if generates plot. default is empty string
    is_discrete - whether the random sample is discrete. NB: this might be overwritten based on bandwidth learned.
        Small bandwidths will always bring out pulse like PDFs. default is False.
    prev_discrete_value_offset - the difference between the threshold and the previous legitimate value. default is 1.
    weight_sum_steps - the number of bins for weighted sum of k-step cutoffs, default is 20
    reverse - for calculating p(X<threshold), i.e., the smaller the measure value the more severe a patient is.
        default is False
    bandwidth - default bandwidth to use if not search bandwidth, default is 1
    kernel - the kernel to use for KDE, default is gaussian.
    optimise_bandwidth - whether to use grid search to find optimal bandwidth for X. default is True
    do_plot - whether to generate plots, default is True
    """

    # estimate density function
    kde, fitted_bandwith = kde_estimate(X, bandwidth=bandwidth, kernel=kernel, optimise_bandwidth=optimise_bandwidth)

    # detect pulse like PDF
    if fitted_bandwith < 0.1:
        # force is_discrete because the bestfitted_bandwith_bw would lead to one
        logging.info(f"FORCED to be discrete because the bandwidth is {fitted_bandwith}")
        is_discrete = True
    elif fitted_bandwith > 0.7:
        # force is_discrete to be false because the fitted_bandwith would lead to one
        logging.info(f"FORCED to be NOT discrete because the bandwidth is {fitted_bandwith}")
        is_discrete = False

    # automatically adjust on boundaries
    _, adjusted_min = search_for_zero_mass_index(kde, low_bound)
    logging.info(f"adjusted min val {adjusted_min}")
    boundary_offset = low_bound - adjusted_min
    # orig_low_bound = low_bound
    low_bound -= boundary_offset
    up_bound += boundary_offset

    # use learned KDE estimator to get probability
    bins = np.linspace(low_bound, up_bound, n_samples)
    kd_vals = kde.score_samples(bins.reshape(-1, 1))  # Get PDF values for each x
    step_width = (up_bound - low_bound) / n_samples  # get the step
    probs = np.exp(kd_vals) * step_width  # get the approximate prob at each point using the integral of the PDF

    if do_plot:
        tidx = get_threshold_index(
            threshold,
            low_bound,
            is_discrete,
            prev_discrete_value_offset,
            step_width,
            boundary_offset=boundary_offset,
        )
        vis_plot(
            X,
            bins,
            kd_vals,
            title=plot_title,
            vlines=[threshold, round(tidx * step_width + low_bound, 2)],
            vline_colors=["r", "b"],
            vline_labels=["Threshold", "Boudary-adjusted"],
            hist=False,
            reverse=reverse,
        )

    # severity quantification
    if reverse:
        s = low_bound
        e = min(threshold, up_bound)
    else:
        s = max(threshold, low_bound)
        e = up_bound

    # 1. binary like multimorbidity num > 3, yes or no
    sq1 = stepped_severity(
        probs,
        s,
        e,
        1,
        low_bound,
        step_width,
        is_discrete,
        boundary_offset,
        prev_val_offset=prev_discrete_value_offset,
        reverse=reverse,
    )
    # 2. stepped quantification that considers higher/lower the value, more severe the patients are
    sqs = stepped_severity(
        probs,
        s,
        e,
        weight_sum_steps,
        low_bound,
        step_width,
        is_discrete,
        boundary_offset,
        prev_val_offset=prev_discrete_value_offset,
        reverse=reverse,
    )

    return {"overall-prob": round(probs.sum(), 4), "one-step": round(sq1, 4), "k-step": round(sqs, 6), "|X|": len(X)}


def stepped_severity(
    probs: np.ndarray,
    s: float,
    e: float,
    steps: int,
    low_bound: float,
    step_width: float,
    is_discrete: bool,
    boundary_offset: float,
    prev_val_offset: float = 1.0,
    reverse: bool = False,
) -> float:
    """
    To quantify severity by considering higher values as more severe.
    This is done by weighted sum by integrating probs from a threshold - s
    """

    bins = np.linspace(s, e, steps)
    # bin_probs = [probs[get_threshold_index(t, low_bound, is_discrete, step_width):].sum()  for t in bins]
    bin_probs = []
    for i, t in enumerate(bins):
        idx1 = get_threshold_index(t, low_bound, is_discrete, prev_val_offset, step_width, boundary_offset)
        if i < len(bins) - 1:
            idx2 = get_threshold_index(
                bins[i + 1], low_bound, is_discrete, prev_val_offset, step_width, boundary_offset
            )
            bin_probs.append(probs[idx1:idx2].sum())
        else:
            (
                bin_probs.append(probs[idx1:].sum())
                if not reverse
                else bin_probs.append(
                    probs[
                        idx1 : get_threshold_index(
                            e, low_bound, is_discrete, prev_val_offset, step_width, boundary_offset
                        )
                    ].sum()
                )
            )
    s = 0

    # Weight functions
    # def weight_function(x: Any) -> Union[int, Any]:
    #     return math.log(x + 2, 2)

    def weight_function(x: Any) -> Union[int, Any]:
        return x + 1

    w = 0
    for i, p in enumerate(bin_probs):
        if not reverse:
            s += weight_function(i) * p
        else:
            s += weight_function(len(bin_probs) - i) * p
        w += weight_function(i)
    return s / w


def db_ineq(di1: dict[str, int], di2: dict[str, int]) -> float:
    """
    quantify inequality
    """
    return di1["k-step"] / di2["k-step"] - 1



from typing import Callable, Literal

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from daindex.core import db_ineq, deterioration_index


def get_random_sample(df: pd.DataFrame, feature: str, feature_gen_fun: Callable = None) -> np.ndarray:
    """
    Extract the column relating to the feature and return it as a numpy array
    """
    if feature_gen_fun:
        X = df.apply(feature_gen_fun, axis=1).to_numpy().reshape(-1, 1)
    else:
        X = df[feature].to_numpy().reshape(-1, 1)
    return X


def compare_two_groups(
    df1: pd.DataFrame,
    df2: pd.DataFrame,
    feature: str,
    cohort_name1: str,
    cohort_name2: str,
    di_label: str,
    threshold: float | list[float],
    bandwidth: float | Literal["scott", "silverman"] = 1.0,
    is_discrete: bool = False,
    optimise_bandwidth: bool = True,
    do_plot: bool = True,
    feature_gen_fun: Callable = None,
    reverse: bool = False,
) -> tuple[dict[str, int], dict[str, int], float]:
    """
    Obtain the inequality quantification
    df1 - data frame of the first group
    df2 - data frame of the second group
    feature - the feature name of the measurement, not used if feature_gen_fun is not none
    cohort_name1 - the name of the first cohort to be displayed on the plot
    cohort_name2 - the name of the second cohort to be displayed on the plot
    di_label - the label of the deterioration index to be displayed on the plot
    threshold - the threshold for abnormality starting point. This could be a list
        if two cohorts require different thresholds. For example, male/female might have
        different normal ranges for some measurements. The first element will be used for the first one.
    """
    X1 = get_random_sample(df1, feature, feature_gen_fun=feature_gen_fun)
    X2 = get_random_sample(df2, feature, feature_gen_fun=feature_gen_fun)
    # it is very important to use the same min/max values as the k-step weighting needs to
    # put the same weight for the same level of deterioration
    min_v = min([np.min(X1), np.min(X2)])
    max_v = max([np.max(X1), np.max(X2)])

    if isinstance(threshold, list):
        threshold1 = threshold[0]
        threshold2 = threshold[1]
    else:
        threshold1 = threshold2 = threshold
    c1_di = deterioration_index(
        X1,
        min_v,
        max_v,
        reverse=reverse,
        threshold=threshold1,
        bandwidth=bandwidth,
        is_discrete=is_discrete,
        plot_title=f"{cohort_name1}| {di_label}",
        optimise_bandwidth=optimise_bandwidth,
        do_plot=do_plot,
    )
    c2_di = deterioration_index(
        X2,
        min_v,
        max_v,
        reverse=reverse,
        threshold=threshold2,
        bandwidth=bandwidth,
        is_discrete=is_discrete,
        plot_title=f"{cohort_name2}| {di_label}",
        optimise_bandwidth=optimise_bandwidth,
        do_plot=do_plot,
    )
    ineq = db_ineq(c1_di, c2_di)
    return c1_di, c2_di, ineq


def area_under_curve(w_data: np.ndarray, decision_boundary: float = 0.5) -> tuple[float, float]:
    """
    calculate the area under curve - do NOT do interpolation
    """
    prev = None
    area = 0.0
    decision_area = 0.0
    n_points = 0
    for r in w_data:
        if prev is not None:
            a = (r[1] + prev[1]) * (r[0] - prev[0]) / 2  # * r[2]
            area += a
            if prev[0] >= decision_boundary:
                decision_area += a
                n_points += 1
        prev = r

    if prev is not None:
        a = (r[1] + prev[1]) * (r[0] - prev[0]) / 2  # * r[2]
        area += a
        if prev[0] >= decision_boundary:
            decision_area += a
            n_points += 1

    return area, decision_area


def vis_DA_indices(data: np.ndarray, label: str) -> tuple[float, float, np.ndarray]:
    """
    plot dot-line for approximating a DA curve
    """
    w_data = data[np.where(data[:, 1] > 0)][:, [0, 2, 1]]
    a, decision_area = area_under_curve(w_data)
    plt.plot(w_data[:, 0], w_data[:, 1], "-")
    plt.plot(w_data[:, 0], w_data[:, 1], "o", label=label)
    return a, decision_area, w_data


def calc_ratios(a1: float, a2: float, da1: float, da2: float) -> dict[str, float | str]:
    ratio = (a2 - a1) / a1
    decision_ratio = ((da2 - da1) / da1) if da1 != 0 else "N/A"
    return {"Full": ratio, "Decision": decision_ratio}


def get_da_curve(
    d1: np.ndarray,
    d2: np.ndarray,
    g1_label: str,
    g2_label: str,
    deterioration_label: str,
    allocation_label: str,
    config: dict = {},
    decision_boundary: float = 0.5,
    verbose: bool = True,
) -> tuple[dict[str, float | str], plt.Figure]:
    """
    do DA curve visualisation
    """
    if "style" in config:
        plt.style.use(config["style"])
    font_size = config["font_size"] if "font_size" in config else 12
    if "fig_size" in config:
        plt.figure(figsize=config["fig_size"], dpi=200)

    # do some clearning: remove those empty points
    d1 = np.delete(d1, np.where(d1[:, 1] == 0), axis=0)
    d2 = np.delete(d2, np.where(d2[:, 1] == 0), axis=0)
    # make two datasets even in terms of max x val
    x_min = min(np.max(d1[:, 0]), np.max(d2[:, 0]))
    d1 = np.delete(d1, np.where(d1[:, 0] > x_min), axis=0)
    d2 = np.delete(d2, np.where(d2[:, 0] > x_min), axis=0)

    # automatically set x/y limits for better viz
    # x_max = max(np.max(d1[:, 0]), np.max(d2[:, 0]))
    y_max = max(np.max(d1[:, 2]), np.max(d2[:, 2]))

    plt.xlim(0, x_min * 1.05)
    plt.ylim(0, y_max * 1.05)

    # do plots
    a1, da1, _ = vis_DA_indices(d1, g1_label)
    a2, da2, _ = vis_DA_indices(d2, g2_label)

    ratios = calc_ratios(a1, a2, da1, da2)

    if verbose:
        print("Ratio\t{0:.6f}\t{1:.6f}\t{2:.2%}".format(a1, a2, ratios["Full"]))
        print("Decision Ratio\t{0:.6f}\t{1:.6f}\t{2:.2%}".format(da1, da2, ratios["Decision"]))

    # figure finishing up
    plt.xlabel(allocation_label, fontsize=font_size)
    plt.ylabel(deterioration_label, fontsize=font_size)

    # plot decision region
    plt.plot([decision_boundary, decision_boundary], [0, 1], "--", lw=0.8, color="g")
    plt.axvspan(decision_boundary, 1, facecolor="b", alpha=0.1)

    plt.legend(fontsize=font_size, loc="best")

    fig = plt.gcf()
    plt.close()

    return ratios, fig


def calc_area(data: np.ndarray) -> tuple[float, float]:
    wd = data[np.where(data[:, 1] > 0)][:, [0, 2, 1]]
    return area_under_curve(wd)


def get_da_ratios(
    d1: np.ndarray, d2: np.ndarray, verbose: bool = False, return_areas: bool = False
) -> dict[str, float | str]:
    """
    Calculate the area under curve ratios
    """
    # do some clearning: remove those empty points
    d1 = np.delete(d1, np.where(d1[:, 1] == 0), axis=0)
    d2 = np.delete(d2, np.where(d2[:, 1] == 0), axis=0)
    # make two datasets even in terms of max x val
    x_min = min(np.max(d1[:, 0]), np.max(d2[:, 0]))
    d1 = np.delete(d1, np.where(d1[:, 0] > x_min), axis=0)
    d2 = np.delete(d2, np.where(d2[:, 0] > x_min), axis=0)

    a1, da1 = calc_area(d1)
    a2, da2 = calc_area(d2)

    ratios = calc_ratios(a1, a2, da1, da2)

    if return_areas:
        ratios.update({"AUC1": a1, "AUC2": a2, "Decision AUC1": da1, "Decision AUC2": da2})

    if verbose:
        print("Ratio\t{0:.6f}\t{1:.6f}\t{2:.2%}".format(a1, a2, ratios["Full"]))
        print("Decision Ratio\t{0:.6f}\t{1:.6f}\t{2:.2%}".format(da1, da2, ratios["Decision"]))

    return ratios
